"""Get machine learning predictions from geodata raster images"""
import os

import structlog.stdlib
from pathlib import Path

from samgis_core.utilities import session_logger
from samgis_web.utilities.constants import SERVICE_NAME


ROOT = Path(globals().get("__file__", "./_")).absolute().parent.parent
PROJECT_ROOT_FOLDER = Path(os.getenv("PROJECT_ROOT_FOLDER", ROOT))

loglevel = os.getenv('LOGLEVEL', 'INFO').upper()
session_logger.setup_logging(log_level=loglevel)
app_logger = structlog.stdlib.get_logger(__name__)



"""lambda helper functions"""
from typing import Dict

from lisa_on_cuda.utils.app_helpers import get_cleaned_input
from samgis_web.web.web_helpers import get_url_tile, get_source_name

from samgis_lisa import app_logger
from samgis_lisa.utilities.type_hints import StringPromptApiRequestBody


def get_parsed_bbox_points_with_string_prompt(request_input: StringPromptApiRequestBody) -> Dict:
    """
        Parse the raw input request into bbox, prompt string and zoom

        Args:
            request_input: input dict

        Returns:
            dict with bounding box, prompt string and zoom
        """

    app_logger.info(f"try to parsing input request: {type(request_input)}, {request_input}...")
    if isinstance(request_input, str):
        app_logger.info(f"string/json input, parsing it to {type(StringPromptApiRequestBody)}...")
        request_input = StringPromptApiRequestBody.model_validate_json(request_input)
        app_logger.info(f"parsed input, now of type {type(request_input)}...")

    bbox = request_input.bbox
    app_logger.debug(f"request bbox: {type(bbox)}, value:{bbox}.")
    ne = bbox.ne
    sw = bbox.sw
    app_logger.debug(f"request ne: {type(ne)}, value:{ne}.")
    app_logger.debug(f"request sw: {type(sw)}, value:{sw}.")
    ne_latlng = [float(ne.lat), float(ne.lng)]
    sw_latlng = [float(sw.lat), float(sw.lng)]
    new_zoom = int(request_input.zoom)
    cleaned_prompt = get_cleaned_input(request_input.string_prompt)

    app_logger.debug(f"bbox => {bbox}.")
    app_logger.debug(f'request_input-prompt cleaned => {cleaned_prompt}.')

    app_logger.info("unpacking elaborated request...")
    return {
        "bbox": [ne_latlng, sw_latlng],
        "prompt": cleaned_prompt,
        "zoom": new_zoom,
        "source": get_url_tile(request_input.source_type),
        "source_name": get_source_name(request_input.source_type)
    }



"""custom type hints"""
from pydantic import BaseModel
from samgis_web.utilities.type_hints import RawBBox


class StringPromptApiRequestBody(BaseModel):
    """Input lambda request validator type (not yet parsed)"""
    id: str = ""
    bbox: RawBBox
    string_prompt: str
    zoom: int | float
    source_type: str = "OpenStreetMap.Mapnik"
    debug: bool = False



from datetime import datetime
from typing import Callable

from samgis_core.utilities.type_hints import LlistFloat, DictStrInt
from samgis_web.io_package.geo_helpers import get_vectorized_raster_as_geojson
from samgis_web.io_package.raster_helpers import write_raster_tiff, write_raster_png
from samgis_web.io_package.tms2geotiff import download_extent
from samgis_web.utilities.constants import DEFAULT_URL_TILES

from samgis_lisa import app_logger
from samgis_lisa.utilities.constants import LISA_INFERENCE_FN


msg_write_tmp_on_disk = "found option to write images and geojson output..."


def load_model_and_inference_fn(inference_function_name_key: str, inference_decorator: Callable = None):
    """
    If missing, instantiate the inference function as reference the inference_function_name_key
    using the global object models_dict

    Args:
        inference_function_name_key: machine learning model name
        inference_decorator: inference decorator like ZeroGPU (e.g. spaces.GPU)

    Returns:
    """
    from lisa_on_cuda.utils import app_helpers
    from samgis_lisa.prediction_api.global_models import models_dict

    if models_dict[inference_function_name_key]["inference"] is None:
        msg = f"missing inference function {inference_function_name_key}, "
        msg += f"instantiating it now"
        if inference_decorator:
            msg += f" using the inference decorator {inference_decorator.__name__}"
        msg += "..."
        app_logger.info(msg)
        parsed_args = app_helpers.parse_args([])
        inference_fn = app_helpers.get_inference_model_by_args(
            parsed_args,
            internal_logger0=app_logger,
            inference_decorator=inference_decorator
        )
        models_dict[inference_function_name_key]["inference"] = inference_fn


def lisa_predict(
        bbox: LlistFloat,
        prompt: str,
        zoom: float,
        inference_function_name_key: str = LISA_INFERENCE_FN,
        source: str = DEFAULT_URL_TILES,
        source_name: str = None,
        inference_decorator: Callable = None
) -> DictStrInt:
    """
    Return predictions as a geojson from a geo-referenced image using the given input prompt.

    1. if necessary instantiate a segment anything machine learning instance model
    2. download a geo-referenced raster image delimited by the coordinates bounding box (bbox)
    3. get a prediction image from the segment anything instance model using the input prompt
    4. get a geo-referenced geojson from the prediction image

    Args:
        bbox: coordinates bounding box
        prompt: machine learning input prompt
        zoom: Level of detail
        inference_function_name_key: machine learning model name
        source: xyz
        source_name: name of tile provider
        inference_decorator: inference decorator like ZeroGPU (spaces.GPU)

    Returns:
        dict containing the output geojson, the geojson shapes number and a machine learning textual output string
    """
    from os import getenv
    from samgis_lisa.prediction_api.global_models import models_dict

    if source_name is None:
        source_name = str(source)

    msg_start = "start lisa inference"
    if inference_decorator:
        msg_start += f", using the inference decorator {inference_decorator.__name__}"
    msg_start += "..."
    app_logger.info(msg_start)
    app_logger.debug(f"type(source):{type(source)}, source:{source},")
    app_logger.debug(f"type(source_name):{type(source_name)}, source_name:{source_name}.")

    load_model_and_inference_fn(inference_function_name_key, inference_decorator=inference_decorator)
    app_logger.debug(f"using a '{inference_function_name_key}' instance model...")
    inference_fn = models_dict[inference_function_name_key]["inference"]
    app_logger.info(f"loaded inference function '{inference_fn.__name__}'.")

    pt0, pt1 = bbox
    app_logger.info(f"tile_source: {source}: downloading geo-referenced raster with bbox {bbox}, zoom {zoom}.")
    img, transform = download_extent(w=pt1[1], s=pt1[0], e=pt0[1], n=pt0[0], zoom=zoom, source=source)
    app_logger.info(
        f"img type {type(img)} with shape/size:{img.size}, transform type: {type(transform)}, transform:{transform}.")
    folder_write_tmp_on_disk = getenv("WRITE_TMP_ON_DISK", "")
    prefix = f"w{pt1[1]},s{pt1[0]},e{pt0[1]},n{pt0[0]}_"
    if bool(folder_write_tmp_on_disk):
        now = datetime.now().strftime('%Y%m%d_%H%M%S')
        app_logger.info(msg_write_tmp_on_disk + f"with coords {prefix}, shape:{img.shape}, {len(img.shape)}.")
        if img.shape and len(img.shape) == 2:
            write_raster_tiff(img, transform, f"{source_name}_{prefix}_{now}_", f"raw_tiff", folder_write_tmp_on_disk)
        if img.shape and len(img.shape) == 3 and img.shape[2] == 3:
            write_raster_png(img, transform, f"{source_name}_{prefix}_{now}_", f"raw_img", folder_write_tmp_on_disk)
    else:
        app_logger.info("keep all temp data in memory...")

    app_logger.info(f"lisa_zero, source_name:{source_name}, source_name type:{type(source_name)}.")
    app_logger.info(f"lisa_zero, prompt type:{type(prompt)}.")
    app_logger.info(f"lisa_zero, prompt:{prompt}.")
    prompt_str = str(prompt)
    app_logger.info(f"lisa_zero, img type:{type(img)}.")
    embedding_key = f"{source_name}_z{zoom}_{prefix}"
    _, mask, output_string = inference_fn(input_str=prompt_str, input_image=img, embedding_key=embedding_key)
    app_logger.info(f"lisa_zero, output_string type:{type(output_string)}.")
    app_logger.info(f"lisa_zero, mask_output type:{type(mask)}.")
    app_logger.info(f"created output_string '{output_string}', preparing conversion to geojson...")
    return {
        "output_string": output_string,
        **get_vectorized_raster_as_geojson(mask, transform)
    }
